#include <stdlib.h>

#include <utility>

#include "bms_parser.h"

#include "bms_util.h"
#include "bms_exception.h"

namespace {
  std::pair<unsigned int, unsigned int> factorization( unsigned int x ) {
    if ( x == 0 || x == 1 ) {
      return std::make_pair( 1, 1 );
    }

    unsigned int quotient = 1;
    unsigned int prime_factors[] = {2, 3, 5, 7};
    for ( unsigned int i = 0; i < sizeof( prime_factors ) / sizeof( unsigned int ); ++i ) {
      while ( x % prime_factors[i] == 0 ) {
        x /= prime_factors[i];
        quotient *= prime_factors[i];
      }
    }
    return std::make_pair( quotient, x );
  }
}


// -- BmsParseFrame ------------------------------------------------------
BmsParseFrame::BmsParseFrame( BmsBms*           master,
                              BmsRandomManager& random_manager,
                              bool              make_syntax_tree ) :
parent_( NULL ),
master_( new BmsParseFrameData( master ) ),
current_( new BmsParseFrameData( &random_manager.GetRootData().GetData() ) ),
current_if_statement_( &random_manager.GetRoot() ),
if_statement_position_( 0 ),
make_syntax_tree_( make_syntax_tree )
{
}


BmsParseFrame::BmsParseFrame( BmsParseFrame*               parent,
                              BmsBms*                      current,
                              BmsRandomIfStatementElement* current_if_statement,
                              bool                         master_take_over ) :
parent_( parent ),
master_( master_take_over ? parent->master_ : NULL ),
current_( new BmsParseFrameData( current, parent->current_->bgm_channel_counter_ ) ),
current_if_statement_( current_if_statement ),
if_statement_position_( 0 ),
make_syntax_tree_( parent->make_syntax_tree_ )
{
}


BmsParseFrame::~BmsParseFrame()
{
  if ( parent_ == NULL ) {
    delete master_;
  }
  delete current_;
}

void
BmsParseFrame::CallWriteDonwFunction( BmsParser& parser, BmsParserWriteDownFunction function )
{
  if ( master_ != NULL ) {
    (parser.*function)( *master_ );
  }
  if ( make_syntax_tree_ ) {
    (parser.*function)( *current_ );
  }
}

// -- BmsParseFrameStack -------------------------------------------------
BmsParseFrameStack::BmsParseFrameStack( BmsBms& bms,
                                        BmsRandomManager& random_manager,
                                        bool make_syntax_tree ) :
stack_(),
random_manager_( random_manager ),
random_value_queue_(),
make_syntax_tree_( make_syntax_tree )
{
  stack_.push( new BmsParseFrame( &bms, random_manager, make_syntax_tree_ ) );
}

BmsParseFrameStack::~BmsParseFrameStack()
{
  while ( NOT( stack_.empty() ) ) {
    BmsParseFrame* tmp = stack_.top();
    delete tmp;
    stack_.pop();
  }
}

void
BmsParseFrameStack::SetRandomValue( unsigned int value )
{
  random_value_queue_.push( value );
}

void
BmsParseFrameStack::PushGenerateRandomElement( unsigned int max )
{
  if ( make_syntax_tree_ ) {
    stack_.top()->current_if_statement_->MakeNewGenerateRandomElement( max );
  }
}

void
BmsParseFrameStack::PushIfStatementElement( unsigned int prerequisite )
{
  if ( make_syntax_tree_ ) {
    BmsRandomIfStatementElement& if_statement = stack_.top()->current_if_statement_->MakeNewIfStatementElement( prerequisite );
    stack_.push( new BmsParseFrame( stack_.top(),
                                    &if_statement.MakeNewDataElement().GetData(),
                                    &if_statement,
                                    prerequisite == random_manager_.GetValue() ) );
  }
  else {
    stack_.push( new BmsParseFrame( stack_.top(), NULL, NULL,
                                    prerequisite == random_manager_.GetValue() ) );
  }
}

void
BmsParseFrameStack::PopIfStatementElement( void )
{
  BmsParseFrame* tmp = stack_.top();
  stack_.pop();
  delete tmp;

  if ( make_syntax_tree_ ) {
    stack_.top()->current_->bms_ = &stack_.top()->current_if_statement_->MakeNewDataElement().GetData();
  }
}

unsigned int
BmsParseFrameStack::ReplayRandomValue( void )
{
  stack_.top()->if_statement_position_ += 1;

  if ( make_syntax_tree_ ) {
    if ( stack_.top()->GetChildElementAs<BmsRandomGenerateRandomElement>() == NULL ) {
      throw BMS_INTERNAL_EXCEPTION;
    }
  }
  unsigned int tmp = random_value_queue_.front();
  random_value_queue_.pop();
  return tmp;
}

void
BmsParseFrameStack::MoveNextChildIfstatementElement( unsigned int prerequisite )
{
  stack_.top()->if_statement_position_ += 1;

  if ( make_syntax_tree_ ) {
    BmsRandomIfStatementElement* child_if_statement = stack_.top()->GetChildElementAs<BmsRandomIfStatementElement>();
    if ( child_if_statement == NULL ) {
      throw BMS_INTERNAL_EXCEPTION;
    }
    BmsRandomDataElement* child_current = dynamic_cast<BmsRandomDataElement*>( child_if_statement->GetChildren()[0] );
    if ( child_current == NULL ) {
      throw BMS_INTERNAL_EXCEPTION;
    }
    if ( child_if_statement->GetPrequisite() != prerequisite ) {
      throw BMS_INTERNAL_EXCEPTION;
    }
    stack_.push( new BmsParseFrame( stack_.top(),
                                    &child_current->GetData(),
                                    child_if_statement,
                                    child_if_statement->GetPrequisite() == random_manager_.GetValue() ) );
  }
  else {
    stack_.push( new BmsParseFrame( stack_.top(), NULL, NULL,
                                    prerequisite == random_manager_.GetValue() ) );
  }
}

void
BmsParseFrameStack::MoveParentIfStatementElement( void )
{
  if ( stack_.size() <= 1 ) {
    throw BMS_INTERNAL_EXCEPTION;
  }
  BmsParseFrame* tmp = stack_.top();
  stack_.pop();
  delete tmp;

  stack_.top()->if_statement_position_ += 1;
  if ( make_syntax_tree_ ) {
    BmsRandomDataElement* next_if_statement = stack_.top()->GetChildElementAs<BmsRandomDataElement>();
    if ( next_if_statement == NULL ) {
      throw BMS_INTERNAL_EXCEPTION;
    }
    stack_.top()->current_->bms_ = &next_if_statement->GetData();
  }
}


unsigned int
BmsParseFrameStack::GetSize( void )
{
  return stack_.size();
}


void
BmsParseFrameStack::CallWriteDonwFunction( BmsParser& parser, BmsParserWriteDownFunction function )
{
  stack_.top()->CallWriteDonwFunction( parser, function );
}


// -- BmsParser ----------------------------------------------------------
BmsParser::BmsParser( void ) :
duplicate_header_handler_( NULL ),
duplicate_bar_change_handler_( NULL ),
parse_error_ignore_handler_( NULL ),
generate_random_handler_( NULL ),
make_syntax_tree_( false ),

ignore_channel_set_(),

is_first_parse_( true ),
line_data_(),
line_number_( 0 ),
random_manager_( NULL ),
frame_stack_(),

bar_( 0 ),
channel_( BmsWord::MIN ),
key_( "" ),
value_( "" )
{
}


void
BmsParser::SetDuplicateHeaderHandler( BmsHandler::DuplicateHeader* handler )
{
  duplicate_header_handler_ = handler;
}

void
BmsParser::SetDuplicateBarChangeHandler( BmsHandler::DuplicateBarChange* handler )
{
  duplicate_bar_change_handler_ = handler;
}

void
BmsParser::SetParseErrorIgnoreHandler( BmsHandler::ParseErrorIgnore* handler )
{
  parse_error_ignore_handler_ = handler;
}

void
BmsParser::SetGenerateRandomHandler( BmsHandler::GenerateRandom* handler )
{
  generate_random_handler_ = handler;
}

void
BmsParser::SetMakeSyntaxTreeFlag( bool make_syntax_tree )
{
  make_syntax_tree_ = make_syntax_tree;
}


void
BmsParser::AddIgnoreChannel( BmsWord channel_number )
{
  ignore_channel_set_.insert( channel_number );
}


unsigned int
BmsParser::GetLineNumber( void ) const
{
  return line_number_;
}


void
BmsParser::Parse( const std::string& filename, BmsBms& bms, BmsRandomManager& random_manager )
{
  BmsTextFileReader reader( filename );
  this->Parse( reader, bms, random_manager );
}

void
BmsParser::Parse( BmsTextFileReader& reader, BmsBms& bms, BmsRandomManager& random_manager )
{
  {
    std::string tmp;
    while ( reader.ReadLine( tmp, true ) ) {
      for ( unsigned int i = tmp.length() - 1; i != 0; --i ) {
        if ( tmp[i] == ' ' || tmp[i] == '\t' ) {
          tmp.erase( i );
        }
        else {
          break;
        }
      }
      line_data_.push_back( tmp );
    }
  }

  is_first_parse_ = true;
  line_number_ = 0;
  random_manager_ = &random_manager;
  BmsParseFrameStack a( bms, random_manager, make_syntax_tree_ );
  frame_stack_ = &a;
  std::set<unsigned int> ignore_line_set;

  for ( std::vector<std::string>::const_iterator it = line_data_.begin(); it != line_data_.end(); ++it ) {
    ++line_number_;
    try {
      this->ParseDefault( it->c_str() );
    }
    catch ( BmsParseException& e ) {
      BmsHandler::ParseErrorIgnore* handler = parse_error_ignore_handler_;
      if ( handler == NULL ) {
        handler = &BmsHandler::ParseErrorIgnore::DefaultObject;
      }
      if ( NOT( handler->Handle( *this, e, *it ) ) ) {
        throw;
      }
      ignore_line_set.insert( line_number_ );
    }
  }
  if ( frame_stack_->GetSize() > 1 ) {
    throw BmsParseNotFoundEndIfException( line_number_ );
  }

  is_first_parse_ = false;
  line_number_ = 0;
  for ( std::vector<std::string>::const_iterator it = line_data_.begin(); it != line_data_.end(); ++it ) {
    ++line_number_;
    if ( ignore_line_set.find( line_number_ ) != ignore_line_set.end() ) {
      continue;
    }
    try {
      this->ParseDefault( it->c_str() );
    }
    catch ( BmsParseException& e ) {
      BmsHandler::ParseErrorIgnore* handler = parse_error_ignore_handler_;
      if ( handler == NULL ) {
        handler = &BmsHandler::ParseErrorIgnore::DefaultObject;
      }
      if ( NOT( handler->Handle( *this, e, *it ) ) ) {
        throw;
      }
    }
  }
  if ( frame_stack_->GetSize() > 1 ) {
    throw BmsParseNotFoundEndIfException( line_number_ );
  }
}

void
BmsParser::ParseDefault( const char* str )
{
  while ( *str == ' ' || *str == '\t' ) {
    ++str;
  }
  switch ( *str ) {
  case '#':
    this->ParseHeaderOrBar( str + 1 );
    break;

  case '\0':
    break;

  default:
    this->ParseComment( str );
    break;
  }
}


void
BmsParser::ParseComment( const char* str )
{
  // ͓ɉȂ
  NOT_USE_VAR( str );
}


void
BmsParser::ParseHeaderOrBar( const char* str )
{
  if ( BmsUtil::IsDigit( *str ) ) {
    this->ParseBar( str );
  }
  else if ( BmsUtil::IsAlphabet( *str ) ) {
    this->ParseHeaderKey( str );
  }
  else {
    throw BmsParseInvalidCharAsHeaderOrBarException( line_number_, std::string( str, 1 ) );
  }
}


void
BmsParser::ParseHeaderKey( const char* str )
{
  key_.clear();
  for ( unsigned int i = 0; ; ++i ) {
    if ( str[i] == ' ' || str[i] == '\t' || str[i] == '\0' ) {
      key_.assign( str, i );
      BmsUtil::StringToUpper( key_ );
      this->ParseHeaderSeparator( str + i );
      break;
    }
    if ( BmsUtil::IsNotHex36( str[i] ) ) {
      throw BmsParseInvalidCharAsHeaderKeyException( line_number_, std::string( str + i, 1 ) );
    }
  }
}


void
BmsParser::ParseHeaderSeparator( const char* str )
{
  while ( *str == ' ' || *str == '\t' ) {
    ++str;
  }
  this->ParseHeaderValue( str );
}


void
BmsParser::ParseHeaderValue( const char* str )
{
  value_.assign( str );

  if ( key_ == "RANDOM" || key_ == "IF" || key_ == "ENDIF" ) {
    this->ParseRandomStatement();
    return;
  }

  if ( is_first_parse_ ) {
    return;
  }

  BmsParserWriteDownFunction function;

  // WAV o^
  if ( BmsRegistArray::CheckConstruction( "WAV", key_ ) ) {
    get_regist_array_function_ = &BmsBms::GetWavArray;
    function = &BmsParser::WriteDownRegistArray;
  }
  // BMP o^
  else if ( BmsRegistArray::CheckConstruction( "BMP", key_ ) ) {
    get_regist_array_function_ = &BmsBms::GetBmpArray;
    function = &BmsParser::WriteDownRegistArray;
  }
  // BPM o^
  else if ( BmsRegistArray::CheckConstruction( "BPM", key_ ) ) {
    get_regist_array_function_ = &BmsBms::GetBpmArray;
    function = &BmsParser::WriteDownRegistArray;
  }
  // STOP o^
  else if ( BmsRegistArray::CheckConstruction( "STOP", key_ ) ) {
    get_regist_array_function_ = &BmsBms::GetStopArray;
    function = &BmsParser::WriteDownRegistArray;
  }
  // ʏwb_o^
  else {
    function = &BmsParser::WriteDownHeader;
  }
  frame_stack_->CallWriteDonwFunction( *this, function );
}


void
BmsParser::WriteDownHeader( BmsParseFrameData& data )
{
  std::string header_value = value_;

  if ( data.bms_->GetHeaders().IsExist( key_ ) ) {
    try {
      if ( duplicate_header_handler_ == NULL ) {
        header_value = BmsHandler::DuplicateHeader::DefaultObject.Handle(
          *this, key_, data.bms_->GetHeaders()[key_], value_ );
      }
      else {
        header_value = duplicate_header_handler_->Handle(
          *this, key_, data.bms_->GetHeaders()[key_], value_ );
      }
    }
    catch ( BmsHandler::DuplicateHeader::Error ) {
      throw BmsParseDuplicateHeaderException( line_number_, key_ );
    }
  }
  data.bms_->GetHeaders().Set( key_, header_value );
}


void
BmsParser::WriteDownRegistArray( BmsParseFrameData& data )
{
  std::string header_value = value_;
  BmsWord pos = BmsWord( key_.substr( key_.length() - 2 ) );
  BmsRegistArray& regist_array = (data.bms_->*get_regist_array_function_)();

  if ( regist_array.IsExist( pos ) ) {
    try {
      if ( duplicate_header_handler_ == NULL ) {
        header_value = BmsHandler::DuplicateHeader::DefaultObject.Handle(
          *this, key_, regist_array[pos], value_ );
      }
      else {
        header_value = duplicate_header_handler_->Handle(
          *this, key_, regist_array[pos], value_ );
      }
    }
    catch ( BmsHandler::DuplicateHeader::Error ) {
      throw BmsParseDuplicateHeaderException( line_number_, key_ );
    }
  }
  regist_array.Set( pos, header_value );
}


void
BmsParser::ParseBar( const char* str )
{
  if ( BmsUtil::IsNotDigit( str[0] ) ||
       BmsUtil::IsNotDigit( str[1] ) ||
       BmsUtil::IsNotDigit( str[2] ) ) {
    throw BmsParseInvalidCharAsBarException( line_number_, std::string( str, 3 ) );
  }
  bar_ = atoi( std::string( str, 3 ).c_str() );
  this->ParseChannel( str + 3 );
}

void
BmsParser::ParseChannel( const char* str )
{
  if ( NOT( BmsWord::CheckConstruction( str ) ) ) {
    throw BmsParseInvalidCharAsChannelException( line_number_, std::string( str, 2 ) );
  }
  channel_ = BmsWord( str );
  if ( ignore_channel_set_.find( channel_ ) != ignore_channel_set_.end() ) {
    return;
  }
  if ( ( channel_ == BmsWord( "02" ) && is_first_parse_ ) ||
       ( channel_ != BmsWord( "02" ) && NOT( is_first_parse_ ) ) ) {
    this->ParseColon( str + 2 );
  }
}

void
BmsParser::ParseColon( const char* str )
{
  if ( *str != ':' ) {
    throw BmsParseInvalidCharAsColonException( line_number_, std::string( str, 1 ) );
  }
  if ( is_first_parse_ ) {
    this->ParseBarChangeValue( str + 1 );
  }
  else {
    this->ParseObjectArray( str + 1 );
  }
}

void
BmsParser::ParseBarChangeValue( const char* str )
{
  if ( NOT( BmsUtil::StringToFloat( str, &bar_change_ratio_ ) ) ) {
    throw BmsParseInvalidBarChangeValueException( line_number_, std::string( str ) );
  }
  if ( bar_change_ratio_ <= 0.0 ) {
    throw BmsParseInvalidBarChangeValueException( line_number_, std::string( str ) );
  }
  frame_stack_->CallWriteDonwFunction( *this, &BmsParser::WriteDownBarChangeValue );
}

void
BmsParser::WriteDownBarChangeValue( BmsParseFrameData& data )
{
  double tmp = bar_change_ratio_;

  if ( data.bms_->GetBarManager()[bar_].GetRatio() != 1.0 ) {
    try {
      if ( duplicate_bar_change_handler_ == NULL ) {
        tmp = BmsHandler::DuplicateBarChange::DefaultObject.Handle(
          *this, bar_, data.bms_->GetBarManager()[bar_].GetRatio(), bar_change_ratio_ );
      }
      else {
        tmp = duplicate_bar_change_handler_->Handle(
          *this, bar_, data.bms_->GetBarManager()[bar_].GetRatio(), bar_change_ratio_ );
      }
    }
    catch ( BmsHandler::DuplicateBarChange::Error ) {
      throw BmsParseDuplicateBarChangeException( line_number_, bar_ );
    }
  }
  unsigned int new_bar_division_count =
    this->DecideBarDivisionCountFromBarChange( data.bms_->GetBarManager().GetBarDivisionCount(), tmp );
  if ( new_bar_division_count > data.bms_->GetBarManager().GetBarDivisionCount() ) {
    data.bms_->MultiplyBarDivisionCount( new_bar_division_count / data.bms_->GetBarManager().GetBarDivisionCount() );
  }
  data.bms_->GetBarManager()[bar_].SetRatio( tmp );
}


void
BmsParser::ParseObjectArray( const char* str )
{
  word_array_.clear();

  if ( *str == '\0' || *str == ' ' || *str == '\t' ) {
    // 擪ŏI̓IuWF
    throw BmsParseNoObjectArrayException( line_number_ );
  }

  for ( unsigned int i = 0; str[i] != '\0'; i += 2 ) {
    if ( BmsWord::CheckConstruction( str + i ) ) {
      word_array_.push_back( BmsWord( str + i ) );
    }
    else if ( str[i] == '\0' || str[i] == ' ' || str[i] == '\t' ) {
      // ̋󔒖
      while ( str[i] != '\0' ) {
        if ( str[i] != ' ' && str[i] != '\t' ) {
          throw BmsParseInvalidEndAsObjectArrayException( line_number_ );
        }
        ++i;
      }
      break;
    }
    else {
      throw BmsParseInvalidCharAsObjectArrayException( line_number_, std::string( str + i, str[i + 1] != '\0' ? 2 : 1 ) );
    }
  }

  frame_stack_->CallWriteDonwFunction( *this, &BmsParser::WriteDownObjectArray );
}


void
BmsParser::WriteDownObjectArray( BmsParseFrameData& data )
{
  unsigned int to = data.bms_->GetBarManager().GetChennelPositionByBarNumber( bar_ );
  unsigned int length = data.bms_->GetBarManager()[bar_].GetLength();

  if ( length % word_array_.size() != 0 ) {
    unsigned int new_bar_division_count = this->DecideBarDivisionCountFromObjectArray(
      data.bms_->GetBarManager().GetBarDivisionCount(), data.bms_->GetBarManager()[bar_].GetRatio() );

    if ( new_bar_division_count > data.bms_->GetBarManager().GetBarDivisionCount() ) {
      data.bms_->MultiplyBarDivisionCount( new_bar_division_count / data.bms_->GetBarManager().GetBarDivisionCount() );
      to = data.bms_->GetBarManager().GetChennelPositionByBarNumber( bar_ );
      length = data.bms_->GetBarManager()[bar_].GetLength();
    }
  }

  unsigned int channel_dup_count;
  if ( channel_ == BmsWord( "01" ) ) {
    channel_dup_count = (*data.bgm_channel_counter_)[bar_];
    ((*data.bgm_channel_counter_)[bar_]) += 1;
  }
  else {
    channel_dup_count = 0;
  }

  int k = 0;
  for ( double i = static_cast<double>( to );
        static_cast<unsigned int>( i ) < to + length;
        i += static_cast<double>( length ) / static_cast<double>( word_array_.size() ) ) {
    data.bms_->GetChannelManager()[channel_][channel_dup_count][static_cast<unsigned int>( i )] = word_array_[k];
    //     BmsChannelManager& mg = bms.GetChannelManager();
    //     BmsChannel& ch = mg[channel_];
    //     BmsChannelBuffer& chbf = ch[channel_dup_count];
    //     BmsWord wd = word_array[static_cast<unsigned int>( i )];
    //     chbf[static_cast<unsigned int>( i )] = wd;
    ++k;
  }
}


void
BmsParser::ParseRandomStatement( void )
{
  if ( is_first_parse_ ) {
    if ( key_ == "RANDOM" ) {
      int tmp;
      if ( NOT( BmsUtil::StringToInteger( value_, &tmp, 10 ) ) ) {
        throw BmsParseInvalidRandomValueException( line_number_, key_, value_ );
      }
      if ( tmp <= 0 ) {
        throw BmsParseInvalidRandomValueException( line_number_, key_, value_ );
      }
      unsigned int random_max = static_cast<unsigned int>( tmp );
      if ( generate_random_handler_ == NULL ) {
        random_manager_->SetValue( BmsHandler::GenerateRandomDefault::DefaultObject.GenerateWithParser( *this, random_max ) );
      }
      else {
        random_manager_->SetValue( generate_random_handler_->GenerateWithParser( *this, random_max ) );
      }
      frame_stack_->SetRandomValue( random_manager_->GetValue() );
      frame_stack_->PushGenerateRandomElement( random_max );
    }
    else if ( key_ == "IF" ) {
      int prerequisite;
      if ( NOT( BmsUtil::StringToInteger( value_, &prerequisite, 10 ) ) ||
           prerequisite <= 0 ) {
        throw BmsParseInvalidRandomValueException( line_number_, key_, value_ );
      }
      frame_stack_->PushIfStatementElement( static_cast<unsigned int>( prerequisite ) );
    }
    else if ( key_ == "ENDIF" ) {
      frame_stack_->PopIfStatementElement();
    }
  }
  else {
    if ( key_ == "RANDOM" ) {
      random_manager_->SetValue( frame_stack_->ReplayRandomValue() );
    }
    else if ( key_ == "IF" ) {
      int prerequisite;
      if ( NOT( BmsUtil::StringToInteger( value_, &prerequisite, 10 ) ) ||
           prerequisite <= 0 ) {
        throw BmsParseInvalidRandomValueException( line_number_, key_, value_ );
      }
      frame_stack_->MoveNextChildIfstatementElement( static_cast<unsigned int>( prerequisite ) );
    }
    else if ( key_ == "ENDIF" ) {
      frame_stack_->MoveParentIfStatementElement();
    }
  }
}

unsigned int
BmsParser::DecideBarDivisionCountFromBarChange( unsigned int current, double bar_change_ratio )
{
  unsigned int request = BmsUtil::GetDenominator( bar_change_ratio, BmsConst::BAR_DIVISION_COUNT_MAX );

  if ( current == 0 || request == 0 ) {
    throw BMS_INTERNAL_EXCEPTION;
  }

  std::pair<unsigned int, unsigned int> factors = factorization( BmsUtil::LCM( current, request ) );
  while ( factors.first < factors.second * 4 ) {
    factors.first *= 2;
  }
  if ( factors.first > BmsConst::BAR_DIVISION_COUNT_MAX ) {
    unsigned int prev = 1;
    for ( unsigned int i = 1; ; ++i ) {
      if ( factorization( i ).second != 1 ) {
        continue;
      }
      if ( current * i > BmsConst::BAR_DIVISION_COUNT_MAX ) {
        return current * prev;
      }
      prev = i;
    }
  }
  return factors.first;
}


unsigned int
BmsParser::DecideBarDivisionCountFromObjectArray( unsigned int current, double bar_change_ratio )
{
  unsigned int request = word_array_.size() * BmsUtil::GetDenominator( bar_change_ratio, BmsConst::BAR_DIVISION_COUNT_MAX );
  if ( current == 0 || request == 0 ) {
    throw BMS_INTERNAL_EXCEPTION;
  }

  std::pair<unsigned int, unsigned int> factors = factorization( BmsUtil::LCM( current, request ) );
  while ( factors.first < factors.second * 4 ) {
    factors.first *= 2;
  }
  if ( factors.first > BmsConst::BAR_DIVISION_COUNT_MAX ) {
    for ( unsigned int i = 1; ; ++i ) {
      if ( factorization( i ).second != 1 ) {
        continue;
      }
      if ( current * i > request * 3 ) {
        return current * i;
      }
      if ( current * i > BmsConst::BAR_DIVISION_COUNT_MAX ) {
        break;
      }
    }
    throw BmsParseTooManyObjectException( line_number_ );
  }
  return factors.first;
}
