#ifndef BMSBEL_PARSER_H_
#define BMSBEL_PARSER_H_

#include <vector>
#include <stack>
#include <queue>
#include <set>

#include "bms_define.h"
#include "bms_bms.h"
#include "bms_random_manager.h"
#include "bms_text_file_reader.h"
#include "bms_handler.h"

class BmsParser;
struct BmsParseFrameData;
typedef void (BmsParser::*(BmsParserWriteDownFunction))( BmsParseFrameData& );


// -- BmsParseFrameData --------------------------------------------------
struct BmsParseFrameData {
public:
  typedef std::vector<unsigned int> BgmChannelCounter;

  explicit BmsParseFrameData( BmsBms* bms ) :
  bms_( bms ), bgm_channel_counter_( new BgmChannelCounter( BmsConst::BAR_MAX_COUNT ) ) {
  }

  explicit BmsParseFrameData( BmsBms* bms, BgmChannelCounter* counter ) :
  bms_( bms ), bgm_channel_counter_( new BgmChannelCounter( BmsConst::BAR_MAX_COUNT ) ) {
    *bgm_channel_counter_ = *counter;
  }

  ~BmsParseFrameData() {
    delete bgm_channel_counter_;
  }

public:
  BmsBms*            bms_;
  BgmChannelCounter* bgm_channel_counter_;
};

// -- BmsParseFrame ------------------------------------------------------
struct BmsParseFrame {
public:

  explicit BmsParseFrame( BmsBms*           master,
                          BmsRandomManager& random_manager,
                          bool              make_syntax_tree );

  explicit BmsParseFrame( BmsParseFrame*               parent,
                          BmsBms*                      current,
                          BmsRandomIfStatementElement* current_if_statement,
                          bool                         master_take_over );

  ~BmsParseFrame();

  template <class TYPE>
  TYPE* GetChildElementAs( void ) {
    return dynamic_cast<TYPE*>( current_if_statement_->GetChildren()[if_statement_position_] );
  }

  void CallWriteDonwFunction( BmsParser& parser, BmsParserWriteDownFunction function );

public:
  BmsParseFrame*               parent_;
  BmsParseFrameData*           master_;
  BmsParseFrameData*           current_;
  BmsRandomIfStatementElement* current_if_statement_;
  unsigned int                 if_statement_position_;
  bool                         make_syntax_tree_;
};


// -- BmsParseFrameStack -------------------------------------------------
class BmsParseFrameStack {
public:
  explicit BmsParseFrameStack( BmsBms& bms,
                               BmsRandomManager& random_manager,
                               bool make_syntax_tree );

  ~BmsParseFrameStack();

  void SetRandomValue( unsigned int value );
  void PushGenerateRandomElement( unsigned int max );
  void PushIfStatementElement( unsigned int prerequisite );
  void PopIfStatementElement( void );

  unsigned int ReplayRandomValue( void );
  void MoveNextChildIfstatementElement( unsigned int prerequisite );
  void MoveParentIfStatementElement( void );

  unsigned int GetSize( void );

  void CallWriteDonwFunction( BmsParser& parser, BmsParserWriteDownFunction function );

private:
  std::stack<BmsParseFrame*> stack_;
  BmsRandomManager&          random_manager_;
  std::queue<unsigned int>   random_value_queue_;
  bool                       make_syntax_tree_;
};

// -- BmsParser ----------------------------------------------------------
class BmsParser {
public:
  explicit BmsParser( void );

  void Parse( const std::string& filename, BmsBms& bms, BmsRandomManager& random_manager );
  void Parse( BmsTextFileReader& reader, BmsBms& bms, BmsRandomManager& random_manager );

  void SetDuplicateHeaderHandler( BmsHandler::DuplicateHeader* handler );
  void SetDuplicateBarChangeHandler( BmsHandler::DuplicateBarChange* handler );
  void SetParseErrorIgnoreHandler( BmsHandler::ParseErrorIgnore* handler );
  void SetGenerateRandomHandler( BmsHandler::GenerateRandom* handler );
  void SetMakeSyntaxTreeFlag( bool make_syntax_tree );

  void AddIgnoreChannel( BmsWord channel_number );

  unsigned int GetLineNumber( void ) const;

private:
  void ParseDefault( const char* str );
  void ParseComment( const char* str );
  void ParseHeaderOrBar( const char* str );
  void ParseHeaderKey( const char* str );
  void ParseHeaderSeparator( const char* str );
  void ParseHeaderValue( const char* str );
  void WriteDownHeader( BmsParseFrameData& data );
  void WriteDownRegistArray( BmsParseFrameData& data );
  void ParseBar( const char* str );
  void ParseChannel( const char* str );
  void ParseColon( const char* str );
  void ParseBarChangeValue( const char* str );
  void WriteDownBarChangeValue( BmsParseFrameData& data );
  void ParseObjectArray( const char* str );
  void WriteDownObjectArray( BmsParseFrameData& data );
  void ParseRandomStatement( void );

  unsigned int DecideBarDivisionCountFromBarChange( unsigned int current, double bar_change_ratio );
  unsigned int DecideBarDivisionCountFromObjectArray( unsigned int current, double bar_change_ratio );

private:
  BmsHandler::DuplicateHeader*    duplicate_header_handler_;
  BmsHandler::DuplicateBarChange* duplicate_bar_change_handler_;
  BmsHandler::ParseErrorIgnore*   parse_error_ignore_handler_;
  BmsHandler::GenerateRandom*     generate_random_handler_;
  bool                            make_syntax_tree_;

  std::set<BmsWord> ignore_channel_set_;

  bool                       is_first_parse_;
  std::vector<std::string>   line_data_;
  unsigned int               line_number_;
  BmsRandomManager*          random_manager_;
  BmsParseFrameStack*        frame_stack_;

  unsigned int              bar_;
  BmsWord                   channel_;
  std::string               key_;
  std::string               value_;
  BmsRegistArray& (BmsBms::*get_regist_array_function_)( void );
  double                    bar_change_ratio_;
  std::vector<BmsWord>      word_array_;
};

#endif // BMSBEL_PARSER_H_
